@@ -1,17 +1,37 @@
+// Copyright (C) 2021  Tassilo Horn <tsdh@gnu.org>
+//
+// This program is free software: you can redistribute it and/or modify it
+// under the terms of the GNU General Public License as published by the Free
+// Software Foundation, either version 3 of the License, or (at your option)
+// any later version.
+//
+// This program is distributed in the hope that it will be useful, but WITHOUT
+// ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
+// FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for
+// more details.
+//
+// You should have received a copy of the GNU General Public License along with
+// this program.  If not, see <https://www.gnu.org/licenses/>.
+
 //! Convenience data structures built from the IPC structs.
 
+use crate::config as cfg;
 use crate::ipc;
+use crate::ipc::NodeMethods;
 use crate::util;
-use ipc::NodeMethods;
 use std::cmp;
 use std::collections::HashMap;
 use std::fmt;
-use swayipc::reply as r;
+use swayipc as s;
+
+pub trait DisplayFormat {
+    fn format_for_display(&self, config: &cfg::Config) -> String;
+}
 
 #[derive(Debug)]
 pub struct Window<'a> {
-    node: &'a r::Node,
-    workspace: &'a r::Node,
+    node: &'a s::Node,
+    workspace: &'a s::Node,
     extra_props: Option<ipc::ExtraProps>,
 }
 
@@ -90,25 +110,63 @@ impl<'a> fmt::Display for Window<'a> {
     fn fmt(&self, f: &mut fmt::Formatter<'_>) -> Result<(), fmt::Error> {
         write!(
             f,
-            "<span font_weight=\"bold\" {}>“{}”</span>   \
-             <i>{}</i>   \
-             on workspace <b>{}</b>   \
-             <span alpha=\"20000\">id {}</span>", // Almost hide ID!
-            if self.node.urgent {
-                " background=\"darkred\" foreground=\"white\""
-            } else {
-                ""
-            },
+            "\"{}\" — {} on workspace {} (id: {}, urgent: {})",
             self.get_title(),
             self.get_app_name(),
             self.workspace.name.as_ref().unwrap(),
-            self.get_id()
+            self.get_id(),
+            self.node.urgent
         )
     }
 }
 
+impl<'a> DisplayFormat for Window<'a> {
+    fn format_for_display(&self, cfg: &cfg::Config) -> String {
+        let default_format = cfg::Format::default();
+        let fmt = cfg
+            .format
+            .as_ref()
+            .and_then(|f| f.window_format.as_ref())
+            .unwrap_or_else(|| default_format.window_format.as_ref().unwrap());
+        let urgency_start = cfg
+            .format
+            .as_ref()
+            .and_then(|f| f.urgency_start.as_ref())
+            .unwrap_or_else(|| default_format.urgency_start.as_ref().unwrap());
+        let urgency_end = cfg
+            .format
+            .as_ref()
+            .and_then(|f| f.urgency_end.as_ref())
+            .unwrap_or_else(|| default_format.urgency_end.as_ref().unwrap());
+
+        fmt.replace("{id}", format!("{}", self.get_id()).as_str())
+            .replace(
+                "{urgency_start}",
+                if self.is_urgent() {
+                    urgency_start.as_str()
+                } else {
+                    ""
+                },
+            )
+            .replace(
+                "{urgency_end}",
+                if self.is_urgent() {
+                    urgency_end.as_str()
+                } else {
+                    ""
+                },
+            )
+            .replace("{app_name}", self.get_app_name())
+            .replace(
+                "{workspace_name}",
+                self.workspace.name.as_ref().unwrap().as_str(),
+            )
+            .replace("{title}", self.get_title())
+    }
+}
+
 fn build_windows<'a>(
-    root: &'a r::Node,
+    root: &'a s::Node,
     include_scratchpad_windows: bool,
     extra_props: Option<&HashMap<i64, ipc::ExtraProps>>,
 ) -> Vec<Window<'a>> {
@@ -130,7 +188,7 @@ fn build_windows<'a>(
 }
 
 fn build_workspaces<'a>(
-    root: &'a r::Node,
+    root: &'a s::Node,
     include_scratchpad: bool,
     extra_props: Option<&HashMap<i64, ipc::ExtraProps>>,
 ) -> Vec<Workspace<'a>> {
@@ -163,7 +221,7 @@ fn build_workspaces<'a>(
 
 /// Gets all application windows of the tree.
 pub fn get_windows<'a>(
-    root: &'a r::Node,
+    root: &'a s::Node,
     include_scratchpad_windows: bool,
     extra_props: Option<&HashMap<i64, ipc::ExtraProps>>,
 ) -> Vec<Window<'a>> {
@@ -177,7 +235,7 @@ pub fn get_windows<'a>(
 
 /// Gets all workspaces of the tree.
 pub fn get_workspaces<'a>(
-    root: &'a r::Node,
+    root: &'a s::Node,
     include_scratchpad: bool,
     extra_props: Option<&HashMap<i64, ipc::ExtraProps>>,
 ) -> Vec<Workspace<'a>> {
@@ -191,14 +249,14 @@ pub fn select_window<'a>(
     prompt: &str,
     windows: &'a [Window],
 ) -> Option<&'a Window<'a>> {
-    util::wofi_select(prompt, windows)
+    util::select_from_choices(prompt, windows)
 }
 
 pub fn select_workspace<'a>(
     prompt: &str,
     workspaces: &'a [Workspace],
 ) -> Option<&'a Workspace<'a>> {
-    util::wofi_select(prompt, workspaces)
+    util::select_from_choices(prompt, workspaces)
 }
 
 pub enum WsOrWin<'a> {
@@ -218,6 +276,17 @@ impl<'a> fmt::Display for WsOrWin<'a> {
     }
 }
 
+impl DisplayFormat for WsOrWin<'_> {
+    fn format_for_display(&self, cfg: &cfg::Config) -> String {
+        match self {
+            WsOrWin::Ws { ws } => ws.format_for_display(cfg),
+            WsOrWin::Win { win } => {
+                "\t".to_owned() + &win.format_for_display(cfg)
+            }
+        }
+    }
+}
+
 impl WsOrWin<'_> {
     pub fn from_workspaces<'a>(
         workspaces: &'a [Workspace],
@@ -237,11 +306,11 @@ pub fn select_workspace_or_window<'a>(
     prompt: &'a str,
     ws_or_wins: &'a [WsOrWin<'a>],
 ) -> Option<&'a WsOrWin<'a>> {
-    util::wofi_select(prompt, ws_or_wins)
+    util::select_from_choices(prompt, ws_or_wins)
 }
 
 pub struct Workspace<'a> {
-    node: &'a r::Node,
+    node: &'a s::Node,
     extra_props: Option<ipc::ExtraProps>,
     pub windows: Vec<Window<'a>>,
 }
@@ -292,12 +361,22 @@ impl PartialOrd for Workspace<'_> {
 
 impl<'a> fmt::Display for Workspace<'a> {
     fn fmt(&self, f: &mut fmt::Formatter<'_>) -> Result<(), fmt::Error> {
-        write!(
-            f,
-            "<span font_weight=\"bold\">“Workspace {}”</span>   \
-             <span alpha=\"20000\">id {}</span>", // Almost hide ID!
-            self.get_name(),
-            self.get_id()
-        )
+        write!(f, "“Workspace {}” (id: {})", self.get_name(), self.get_id())
+    }
+}
+
+impl<'a> DisplayFormat for Workspace<'a> {
+    fn format_for_display(&self, cfg: &cfg::Config) -> String {
+        let default_format = cfg::Format::default();
+        let fmt = cfg
+            .format
+            .as_ref()
+            .and_then(|f| f.workspace_format.as_ref())
+            .unwrap_or_else(|| {
+                default_format.workspace_format.as_ref().unwrap()
+            });
+
+        fmt.replace("{id}", format!("{}", self.get_id()).as_str())
+            .replace("{name}", self.get_name())
     }
 }
